// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.apps;

import java.io.*;
import java.util.*;

import w3c.www.http.*;
import w3c.www.mime.*;
import w3c.jigsaw.http.*;
import w3c.jigsaw.html.*;
import w3c.jigsaw.forms.*;
import w3c.jigsaw.resources.* ;

import cryptiX.utils.Name;
import cryptiX.rsa.*;
import cryptiX.user.*;
import cryptiX.asn1.CodingException;

/**
 * This class is written for Jigsaw, the WWW Server from the W3C.
 * You can find more information about Jigsaw at: http://www.w3.org/pub/WWW/Jigsaw.
 *
 * Using this resource you can receive a certificate request generated from Netscape,
 * create a new cert and send this client cert back to Netscape.
 * GetKey.html is a sample form for using this resource.
 */

public class BrowserKeySign extends PostableResource {

    /**
     * Attribute index - The CA certificate file.
     */
    protected static int ATTR_CA_CERTIFICATE = -1;
    /**
     * Attribute index - The CA private key file.
     */
    protected static int ATTR_CA_PRIVATE_KEY = -1 ;
    /**
     * Attribute index - The log file for signed certificates.
     */
    protected static int ATTR_CERT_LOG_FILE = -1 ;
    /**
     * Attribute index - The serial number of the next certificate.
     */
    protected static int ATTR_SERIALNUMBER = -1 ;
    /**
     * Attribute index - The days the certificate is valid from today.
     */
    protected static int ATTR_DAYSVALID = -1 ;


    static {
    	Attribute a   = null ;
	    Class     cls = null ;
    	try {
	        cls = Class.forName("cryptiX.apps.BrowserKeySign") ;
    	} catch (Exception ex) {
	        ex.printStackTrace() ;
	        System.exit(1) ;
    	}
	    // The CA certificate attribute:
    	a = new StringAttribute("CA certificate"
	    			, null
		    		, Attribute.EDITABLE);
    	ATTR_CA_CERTIFICATE = AttributeRegistry.registerAttribute(cls, a);
	    // The CA private key attribute:
    	a = new StringAttribute("CA private key"
				     , null
				     , Attribute.EDITABLE);
    	ATTR_CA_PRIVATE_KEY = AttributeRegistry.registerAttribute(cls, a) ;
	    // The log file attribute:
    	a = new StringAttribute("log file for certificates"
	    			, null
		    		, Attribute.EDITABLE);
    	ATTR_CERT_LOG_FILE = AttributeRegistry.registerAttribute(cls, a);
	    // The days valid attribute:
    	a = new IntegerAttribute("valid from today (months)"
	    			, null
		    		, Attribute.EDITABLE);
    	ATTR_DAYSVALID = AttributeRegistry.registerAttribute(cls, a);
	    // The serial number attribute:
    	a = new StringAttribute("serial number of next certificate"
	    			, null
		    		, Attribute.EDITABLE);
    	ATTR_SERIALNUMBER = AttributeRegistry.registerAttribute(cls, a);
    }

    /**
     * Get the CA Certificate file name.
     */
    public String getCACertificate() {
       	return getString(ATTR_CA_CERTIFICATE, null);
    }

    /**
     * Get the CA private key file name.
     */
    public String getCAPrivateKey() {
      	return getString(ATTR_CA_PRIVATE_KEY, null);
    }

    /**
     * Get the log file for signed certificates.
     */
    public String getLogFile() {
       	return getString(ATTR_CERT_LOG_FILE, null);
    }

    /**
     * Get the number of days till the cert expires.
     */
    public int getMonthsValid() {
       	return getInt(ATTR_DAYSVALID, 180);
    }

    /**
     * Get the serial number of next certificate.
     */
    public String getSerialNumber() {
       	return getString(ATTR_SERIALNUMBER, null);
    }

    /**
     * Sets the serial number of next certificate.
     */
    public void setSerialNumber(String sn) {
       	setString(ATTR_SERIALNUMBER, sn);
    }

    /**
     * handles the reply for this POST
     */
	public Reply handle(Request request, URLDecoder data) throws HTTPException {

                    // HtmlGenerator is used for generating error messages
        HtmlGenerator g = new HtmlGenerator("Certificate Request Error");   // set title
        BigInt.setOutputRadix(16);                      // show BigInts with base 16
        boolean error = false;
        String errorMsg = new String();
        Date date = new Date();
        X509Certificate cert = null;                    // the new cert
        BigInt sn = BigInt.ZERO;

        g.append("<BODY TEXT=\"#000000\" BGCOLOR=\"#C0C0C0\" LINK=\"#0000EE\" VLINK=\"#551A8B\" ALINK=\"#FF0000\">");
        g.append("<H1>Error processing your certificate request!</H1>");
        g.append("<BR>");

            // parse the values from the form
        String CN = data.getValue("commonName");
        if (CN.length() == 0) {
            error = true;
            g.append("<H3>Common Name is a required field!</H3>");
        }
        String OU = data.getValue("organiziationUnit");
        String O = data.getValue("organiziation");
        if (O.length() == 0) {
            error = true;
            g.append("<H3>Organiziation is a required field!</H3>");
        }
        String L = data.getValue("locality");
        if (L.length() == 0) {
            error = true;
            g.append("<H3>Locality is a required field!</H3>");
        }
        String SP = data.getValue("stateOrProvince");
        String C = data.getValue("country");
        if (C.length() == 0) {
            error = true;
            g.append("<H3>Country is a required field!</H3>");
        }

        Name subject = new Name(C, SP, L, O, OU, CN);

        String pk = data.getValue("public_key");
        byte spkacPEM[] = new byte[pk.length()] ;       // SignedPublicKeyAndChallenge in PEM format
	    pk.getBytes (0, spkacPEM.length, spkacPEM, 0) ;


        NetscapeCertRequest ncr = null;

        try {
            ncr = new NetscapeCertRequest(spkacPEM);
            if (!ncr.verify())
                throw new CodingException("Signature not OK!");
        } catch (CodingException ex) {
            errorMsg = errorMsg.concat("NetscapeCertRequest error: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            g.append("<H3>NetscapeCertRequest error: ",ex.getMessage(),"</H3>");
            error = true;
        }

        String certFile = getCACertificate();
        X509Certificate issuerCert = null;
        try {
            issuerCert = new X509Certificate(certFile);
        } catch (CodingException ex) {
            errorMsg = errorMsg.concat("Error decoding issuer cert: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            g.append("<H3>Error decoding issuer cert: ",ex.getMessage(),"</H3>");
            g.append("<H3>The server is misconfigured!</H3>");
            error = true;
        } catch (IOException ex) {
            errorMsg = errorMsg.concat("Error loading issuer cert: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            g.append("<H3>Error loading issuer cert: ",ex.getMessage(),"</H3>");
            g.append("<H3>The server is misconfigured!</H3>");
            error = true;
        }


        String skFile = getCAPrivateKey();
        RSAPrivateKey sk = null;
        try {
            sk = new RSAPrivateKey(skFile);
        } catch (IOException ex) {
            errorMsg = errorMsg.concat("Error loading private key file: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            g.append("<H3>Error loading private key file: ",ex.getMessage(),"</H3>");
            g.append("<H3>The server is misconfigured!</H3>");
            error = true;
        }

        if (!error) {
            cert = new X509Certificate();

            sn = BigInt.valueOf(getSerialNumber(),16);
            cert.setSerialNumber(sn);
            setSerialNumber(sn.add(BigInt.ONE).toString(16));

            cert.setSignatureAlgorithmID(RSAFunctions.md5WithRSAEncryption);
            cert.setIssuer(issuerCert.subject());           // set issuer to subject of issuer cert
            cert.setNotBefore(new Date());                  // not before today

            int months = getMonthsValid();

            date.setMonth(date.getMonth()+months);              // today + x months
            date.toString();                                    // calculate valid date
            cert.setNotAfter(date);
            cert.setSubject(subject);
            cert.setPublicKey(ncr.getPublicKey());

            try {
                cert.signCertificate(sk);
            } catch (CodingException ex) {
            }
        }

                        // use a simple log file
        String logFile = getLogFile();
        try {
            RandomAccessFile file = new RandomAccessFile(logFile, "rw");

            file.seek(file.length());
            if (error) {
                file.writeBytes(errorMsg);
            } else {
                String log = new String("Certificate from "+CN+" signed. Valid from ");
                log = log.concat(new Date()+" to "+date+". SN = 0x"+sn.toString(16));
                file.writeBytes(log);
                file.writeBytes("\n");
                file.writeBytes(cert.publicKey().toString());
            }

            file.writeBytes("\n");
            file.close();
        } catch (IOException ex) {
            System.out.println("Can't open log file: "+logFile);
        }

    	MimeType APPLICATION_X_X509_USER_CERT = null;
	    try {
	        APPLICATION_X_X509_USER_CERT = new MimeType("application/x-x509-user-cert") ;
    	} catch (MimeTypeFormatException e) {
	        System.out.println ("NetscapeKeySign: invalid mime type.") ;
    	}

        Reply reply = request.makeReply(HTTP.OK);
        reply.setContentType(APPLICATION_X_X509_USER_CERT);
        if (error)
            reply.setStream(g);
        else
            reply.setStream(new ByteArrayInputStream (cert.toDERArray()));
        return reply;
	}

	public BrowserKeySign() {
	}
}
