// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

/**
 * This class shows some examples for using the cryptiX library.
 */

package cryptiX.apps;

import java.io.*;
import java.util.*;

import cryptiX.utils.*;
import cryptiX.rsa.*;
import cryptiX.user.*;
import cryptiX.crypto.Md5Hash;
import cryptiX.asn1.*;

class TestCryptiX implements Observer {

    public static void testNetscapeCertRequest(String fileName) {

        NetscapeCertRequest ncr = null;

        try {
            ncr = new NetscapeCertRequest(fileName);
        } catch (CodingException ex) {
            System.out.println("Error decoding NetscapeCertRequest: "+ex.getMessage());
            return;
        } catch (IOException ex) {
            System.out.println("Error loading file: "+fileName+"; "+ex.getMessage());
            return;
        }

        try {
            if (ncr.verify() == true) {
                System.out.println("Signatur OK!");
                System.out.println("Public Key:\n"+ncr.getPublicKey());
            }
            else
                System.out.println("Signatur nicht OK!");
        } catch (CodingException ex) {
            System.out.println("Error checking NetscapeCertRequest: "+ex.getMessage());
        }
    }

    public static void verifySingedCert(String issuerCertFile, String signedCertFile) {

        X509Certificate issuerCert = null;
        X509Certificate signedCert = null;

        try {
                    // get the issuer certificate
            issuerCert = new X509Certificate(issuerCertFile);
                    // get the signed certificate
            signedCert = new X509Certificate(signedCertFile);
        } catch (CodingException ex) {
            System.out.println("Error decoding certificate: "+ex.getMessage());
            return;
        } catch (IOException ex) {
            System.out.println("Error loading file: "+ex.getMessage());
            return;
        }

        try {
            if (signedCert.verify(issuerCert.publicKey()) == true)
                System.out.println("Signed Certificate OK!");
            else
                System.out.println("Signed Certificate NOT OK!");
        } catch (CodingException ex) {
            System.out.println("Error verifying certificate: "+ex.getMessage());
        }
    }

    public static void verifySelfSignedCert(String selfSignedCertFile) {

        X509Certificate selfSignedCert = null;

        try {
                    // get the self signed certificate
            selfSignedCert = new X509Certificate(selfSignedCertFile);
        } catch (CodingException ex) {
            System.out.println("Error decoding certificate: "+ex.getMessage());
            return;
        } catch (IOException ex) {
            System.out.println("Error loading file: "+ex.getMessage());
            return;
        }

        try {
            if (selfSignedCert.verify() == true)
                System.out.println("Selfsigned certificate OK!");
            else
                System.out.println("Selfsigned certificate NOT OK!");

        } catch (CodingException ex) {
            System.out.println("Error verifying certificate: "+ex.getMessage());
        }
    }

    public static void createSelfSignedCert(Name subject, String privateKeyFile, String certFile) {

        RSAPrivateKey privateKey = null;
        try {
           privateKey = new RSAPrivateKey(privateKeyFile);
        } catch (CodingException ex) {
            System.out.println("Error decoding private key: "+ex.getMessage());
            return;
        } catch (IOException ex) {
            System.out.println("Error loading file: "+ex.getMessage());
            return;
        }

        X509Certificate cert = new X509Certificate();

        cert.setSerialNumber(BigInt.valueOf("1000"));
        cert.setSignatureAlgorithmID(RSAFunctions.md5WithRSAEncryption);
        cert.setIssuer(subject);

        Date now = new Date();
        now.toString();
        cert.setNotBefore(now);                         // not before now

        Date date = new Date();
        date.setMonth(date.getMonth()+12);              // today + x months
        date.toString();                                // calculate valid date
        cert.setNotAfter(date);
        cert.setSubject(subject);
        cert.setPublicKey(privateKey.publicKey());

        try {
            cert.signCertificate(privateKey);
        } catch (CodingException ex) {
        }

        cert.setSaveFormat(ASN1.DER);

        try {
            cert.saveToFile(certFile);
        } catch (IOException ex) {
            System.out.println("Error saving file: "+ex.getMessage());
            return;
        }

        System.out.println("Selfsigned cert saved to: "+certFile);
    }

    public static void createSignedCert(String issuerCertFile, String issuerKeyFile, String subjectKeyFile, Name subject, String certFile) {

        String errorMsg = new String();
        X509Certificate cert = null;
        X509Certificate issuerCert = null;
        boolean error = false;

        try {
            issuerCert = new X509Certificate(issuerCertFile);
        } catch (CodingException ex) {
            errorMsg = errorMsg.concat("Error decoding issuer cert: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            error = true;
        } catch (IOException ex) {
            errorMsg = errorMsg.concat("Error loading issuer cert: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            error = true;
        }

        RSAPrivateKey issuerSK = null;
        try {
            issuerSK = new RSAPrivateKey(issuerKeyFile);
        } catch (IOException ex) {
            errorMsg = errorMsg.concat("Error loading issuer private key file: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            error = true;
        }

        RSAPrivateKey subjectSK = null;
        try {
            subjectSK = new RSAPrivateKey(subjectKeyFile);
        } catch (IOException ex) {
            errorMsg = errorMsg.concat("Error loading subject private key file: ");
            errorMsg = errorMsg.concat(ex.getMessage());
            errorMsg = errorMsg.concat("\n");
            error = true;
        }

        if (!error) {
            cert = new X509Certificate();

            BigInt sn = BigInt.valueOf("AC100",16);
            cert.setSerialNumber(sn);

            cert.setSignatureAlgorithmID(RSAFunctions.md5WithRSAEncryption);
            cert.setIssuer(issuerCert.subject());           // set issuer to subject of issuer cert

            Date now = new Date();
            cert.setNotBefore(now);                         // not before now

            Date date = new Date();
            date.setMonth(date.getMonth()+6);               // today + 6 months
            date.toString();                                // calculate valid date
            cert.setNotAfter(date);

            cert.setSubject(subject);
            cert.setPublicKey(subjectSK.publicKey());

            try {
                cert.signCertificate(issuerSK);
            } catch (CodingException ex) {
            }

            cert.setSaveFormat(ASN1.PEM);

            try {
                cert.saveToFile(certFile);
            } catch (IOException ex) {
                System.out.println("Error saving file: "+ex.getMessage());
                return;
            }

            System.out.println("Signed certificate saved to: "+certFile);
        }
        else
            System.out.println("Error: "+errorMsg);
    }


    public void update(Observable o, Object arg) {
        if (arg != null) {
            System.out.println();
            System.out.println((String)arg);
        }
        else {
            System.out.print(".");
            System.out.flush();
        }
    }

    public static void loadKey(String keyFile, String password) {

        try {
            RSAPrivateKey sk = new RSAPrivateKey(keyFile);

            if (sk.isEncrypted()) {
                System.out.println("Private Key is encrypted with "+ASN1.Oid2Name(sk.encryptionAlgorithm()));

                if (password != null)
                    sk.decrypt(password);
                else
                    System.out.println("Password missing!");
            }
            else
                System.out.println("Private Key is not encrypted!");

            System.out.println("Private Key decrypted!");
        } catch (CodingException ex) {
            System.out.println("Wrong password!");
        } catch (IOException ex) {
            System.out.println("Error open file: "+ex.getMessage());
        }
    }


    public void createKey(String keyFile, int bits, String password) {

        RSAPrivateKey sk = new RSAPrivateKey();
        sk.addObserver(this);
        sk.generateKeys(bits);

        sk.setSaveFormat(ASN1.DER);

        if (password != null) {
            sk.setEncryptionAlgorithm(RSAFunctions.pbeWithMD5AndDES_CBC);
            try {
                sk.encrypt(password);
            } catch (CodingException ex) {
            }
        }
        try {
            sk.saveToFile(keyFile);
        } catch (IOException ex) {
            System.out.println("Error saving file: "+ex.getMessage());
            return;
        }
        System.out.println("Private Key saved to file: "+keyFile);
    }

    public static void testCertRequest() {

        X509CertRequest req = null;

        try {
            req = new X509CertRequest("C:\\wwwkey.req");
        } catch (CodingException ex) {
            System.out.println("Error decoding certificate request: "+ex.getMessage());
            return;
        } catch (IOException ex) {
            System.out.println("Error loading file: "+ex.getMessage());
            return;
        }

        try {
            if (req.verify() == true)
                System.out.println("Self Signed Certificate Request OK!");
            else
                System.out.println("Self Signed Certificate Request NOT OK!");

        } catch (CodingException ex) {
            System.out.println("Error verifying certificate request: "+ex.getMessage());
        }

        System.out.println(req);
    }

    public static void testKeyGenSpeed(int bits) {
        long start;
        long end;
        int count = 50;

        double time = 0.0;
        RSAPrivateKey sk = new RSAPrivateKey();

        for (int i=0; i<count; i++) {
            start = System.currentTimeMillis();
            sk.generateKeys(bits);
            end = System.currentTimeMillis();

            time += (double)(end-start) / 1000;
            System.out.println(i+": "+(double)(end-start) / 1000);
        }

        System.out.println("\naverage time for generating a "+bits+" bit key: "+time/count);
    }

    public static void main(String args[]) throws IOException {

        DataInputStream dis = new DataInputStream(System.in);
        String line;
        BigInt.setOutputRadix(16);

        new TestCryptiX().createKey("IssuerKey.der", 512, null);

        new TestCryptiX().createKey("SubjectKey.der", 512, null);

        Name subject = new Name("AT", "", "", "cryptiX Library", "My Toplevel CA", "");
        createSelfSignedCert(subject, "IssuerKey.der", "selfsignedcert.der");

        subject = new Name("AT", "", "", "UT Graz", "IAIK", "Jo Test");
        createSignedCert("selfsignedcert.der", "IssuerKey.der", "SubjectKey.der", subject, "signedcert.der");

        verifySelfSignedCert("selfsignedcert.der");

        verifySingedCert("selfsignedcert.der", "signedcert.der");

        new TestCryptiX().createKey("encryptedKey.der", 512, "Password");

        loadKey("encryptedKey.der", "Password");

        line = dis.readLine();
    }
}

