//
// BerCoder.java --
//
// Author: Pekka Nikander <pnr@tequila.nixu.fi>
//
// Copyright 1996 by Nixu Oy, Helsinki, Finland
//
// Created: Tue Mar 19 13:11:10 1996
// Last modified: Sun Mar 31 22:51:09 1996 pnr
//
// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.asn1;

import java.util.StringTokenizer;

import cryptiX.asn1.ASN;
import cryptiX.asn1.ASN1Object;
import cryptiX.asn1.CodingException;
import cryptiX.rsa.*;

/**
 * The BerCoder objects are strategy objects that are able to convert
 * ASN.1 objects into BER encoded byte arrays and vice versa.
 *
 */

final class BerCoder {

  public BerCoder() {
  }

  /****************************************************************
   * Coder interface
   ****************************************************************/

  /**
   * Initial encoding buffer size
   */
  private static final int BUFFER_SIZE = 8192;

  /**
   * Returns the ASN.1 BER encoded version of the given ASN.1
   * object as a byte array.
   */
  /*
   * The encoding algorithm encodes the values in tail recursive
   * order, starting from the last lower level object.  The
   * encoded values are stored <EM>backwards</EM> to the
   * coding buffer.  In this way we can calculate the encoded
   * data lengths accurately and "for free".
   */
  public byte[] encode(ASN1Object object) throws CodingException {
    int start;
    byte[] buffer = new byte[BUFFER_SIZE];

    // Lock the object for the duration of encoding
    synchronized (object) {
    try {
      start = encode(object, buffer, buffer.length);
    } catch (ArrayIndexOutOfBoundsException e) {
      throw new CodingException("ASN.1 object too big");
    }
    }

    byte[] coding = new byte[buffer.length - start];
    System.arraycopy(buffer, start, coding, 0, buffer.length - start);
    return coding;
  }

  /**
   * Encodes the given ASN.1 subtree <EM>backwards</EM> to the
   * coding buffer, starting at the given location.  Calls
   * type specific subroutines, which in turn may call
   * recursively call this routine.
   *
   * @param object the ASN.1 subtree to be encoded
   * @param coding the coding buffer
   * @param index index to the first occupied location in the buffer
   * @return index to the first byte in new encoding
   */
  int encode(ASN1Object object, byte[] coding, int index) throws CodingException {

    switch (object.asnType.tag) {

    // Generic ASN.1
    case 1:  return     encodeBoolean           (object, coding, index);
    case 2:  return     encodeBigInt            (object, coding, index);
    case 5:  return     encodeNull              (object, coding, index);
    case 6:  return     encodeOid               (object, coding, index);
    case 0x30: return   encodeSequence          ((SEQUENCE)object, coding, index);
      // by Wolfgang Platzer
    case 3:  return     encodeBitString         (object, coding, index);
    case 4:  return     encodeOctetString       (object, coding, index);
    case 19: return     encodePrintableString   (object, coding, index);
    case 20: return     encodeT61String         (object, coding, index);
    case 22: return     encodeIA5String         (object, coding, index);
    case 23: return     encodeUTCTime           (object, coding, index);
    case 0x31: return   encodeSet               ((SEQUENCE)object, coding, index);
    case 0x80: return   encodeContextSpecific   ((CONTEXT)object, coding, index);

    case 7:  throw new InternalError("ObjectDesc coding not implemented");
    case 8:  throw new InternalError("External ASN.1 data not supported");
    case 9:  throw new InternalError("Real coding not implemented");
    case 10: throw new InternalError("Enumerated coding not implemented");
    }
    throw new CodingException("Unknown ASN.1 type " + object.asnType.tag);
  }

  /****************************************************************
   * Encoding primitives
   ****************************************************************/

  /**
   * Encodes the length.
   */
  int encodeLength(int length, byte[] coding, int index) {
    int end = index;
    if (length >= 0 && length <= 127)
      coding[--index] = (byte)length;  // Short definite form
    else {
      while (length != 0) {            // Long definite form
	coding[--index] = (byte)length;// K octets
	length >>>= 8;
      }
      coding[--index] = (byte)((end - index - 1) | 0x80); // K | 0x80
    }
    return index;
  }

  /**
   * Encodes a boolean
   */
  int encodeBoolean(ASN1Object object, byte[] coding, int index) {
    boolean v = ((Boolean)(object.getValueInternal())).booleanValue();

    coding[--index] = (byte)(v?1:0);		// Value
    coding[--index] = (byte)1;      		// Length
    coding[--index] = (byte)object.asnType.tag;	// Tag
    return index;
  }

  /**
   * Encodes an integer.
   */
  int encodeInteger(ASN1Object object, byte[] coding, int index) {
    int end = index;
    int v = ((Integer)(object.getValueInternal())).intValue();

    do {
      coding[--index] = (byte)v;
      v >>= 8;
    } while (v != 0 && v != -1); 		// Until just sign

    // Check if the sign of the end byte and value are different
    if (v == 0 && coding[index] < 0 || v == -1 && coding[index] >= 0)
      coding[--index] = (byte)v; 		// Sign resolver byte

    coding[--index] = (byte)(end - index - 1);	// Length: 1-4
    coding[--index] = (byte)object.asnType.tag; // Tag
    return index;
  }

  /** WP
   * Encodes an BigInt
   */
  int encodeBigInt(ASN1Object object, byte[] coding, int index) {
    int length;
    int end = index;
    BigInt x = new BigInt(12);
    BigInt v = (BigInt)(object.getValueInternal());
    byte[] data = v.toByteArray();

    if ((data[0] & 0x80) > 0) {             // add 0 byte, because first bit is sign
                                            // and number is not negative !!!
        byte[] tmp = new byte[data.length+1];

        System.arraycopy(data, 0, tmp, 1, data.length);
        tmp[0] = 0;
        data = tmp;
    }

    length = data.length;
    index = index - length;
    System.arraycopy(data, 0, coding, index, length);

    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag

    return index;
  }

  /** WP
   * Encodes an BitString
   */
  int encodeBitString(ASN1Object object, byte[] coding, int index) {
    int length;
    int end = index;
    byte[] data = (byte[])(object.getValueInternal());

    length = data.length;
    index = index - length;
    System.arraycopy(data, 0, coding, index, length);

    coding[--index] = 0;                        // 0 bits padding
    length++;
    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag

    return index;
  }

  /** WP
   * Encodes an OctetString
   */
  int encodeOctetString(ASN1Object object, byte[] coding, int index) {
    int length;
    int end = index;
    byte[] data = (byte[])(object.getValueInternal());

    length = data.length;
    index = index - length;
    System.arraycopy(data, 0, coding, index, length);

    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag

    return index;
  }

  /**
   * Encodes an unsigned.
   */
  int encodeUnsigned(ASN1Object object, byte[] coding, int index) {
    int end = index;
    int v = ((Integer)(object.getValueInternal())).intValue();

    do {
      coding[--index] = (byte)v;
      v >>>= 8;
    } while (v != 0);

    if (coding[index] < 0)
      coding[--index] = (byte)v; 		// Sign resolver byte

    coding[--index] = (byte)(end - index - 1);	// Length: 1-5
    coding[--index] = (byte)object.asnType.tag; // Tag
    return index;
  }

  /**
   * Encodes a String
   */
  int encodeString(ASN1Object object, byte[] coding, int index) {
    int length;
    String v = (String)(object.getValueInternal());

    length = v.length();
    index = index - length;
    v.getBytes(0, v.length(), coding, index);
    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag
    return index;
  }

  /**WP
   * Encodes a PrintableString
   */
  int encodePrintableString(ASN1Object object, byte[] coding, int index) {
    int length;
    String v = (String)(object.getValueInternal());

    length = v.length();
    index = index - length;
    v.getBytes(0, v.length(), coding, index);
    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag
    return index;
  }

  /**WP
   * Encodes a T61String
   */
  int encodeT61String(ASN1Object object, byte[] coding, int index) {
    int length;
    String v = (String)(object.getValueInternal());

    length = v.length();
    index = index - length;
    v.getBytes(0, v.length(), coding, index);
    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag
    return index;
  }

  /**WP
   * Encodes a IA5String
   */
  int encodeIA5String(ASN1Object object, byte[] coding, int index) {
    int length;
    String v = (String)(object.getValueInternal());

    length = v.length();
    index = index - length;
    v.getBytes(0, v.length(), coding, index);
    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag
    return index;
  }

  /**WP
   * Encodes a UTCTime
   */
  int encodeUTCTime(ASN1Object object, byte[] coding, int index) {
    int length;
    String v = (String)(object.getValueInternal());

    length = v.length();
    index = index - length;
    v.getBytes(0, v.length(), coding, index);
    index = encodeLength(length, coding, index);
    coding[--index] = (byte)object.asnType.tag; // Tag
    return index;
  }

  /**
   * Encodes a Null
   */
  int encodeNull(ASN1Object object, byte[] coding, int index) {
    coding[--index] = 0;				// Length
    coding[--index] = (byte)object.asnType.tag;   // Tag
    return index;
  }

  /**changed by WP
   * Encodes an OID.  Oids are represented as dotted strings.
   */
  int encodeOid(ASN1Object object, byte[] coding, int index)
    throws CodingException {
    int end = index;
    String v = (String)(object.getValueInternal());
    String[] subid;		// Table of subid values as strings
    StringTokenizer st;		// Stringtokenizer to pick up subids

    st = new StringTokenizer(v, ".");
    subid = new String[st.countTokens()];
    for (int i = 0; i < subid.length; i++)
      subid[i] = st.nextToken();

    for (int i = subid.length-1; i > 1; i--) { // First 2 special, therefore>1
      long si = oidSubId(subid, i);

      coding[--index] = (byte)(si & 0x7f);
      si >>>= 7;
      while (si > 0) {
	coding[--index] = (byte)(si | 0x80);
	si >>>= 7;
      }
    }
    // First two
    if (subid.length > 0) {
      coding[--index] = (byte)(oidSubId(subid, 0) * 40);
      coding[index]  += (byte)oidSubId(subid, 1);
    }
    // Length
    index = encodeLength(end - index, coding, index);
    // Tag
    coding[--index] = (byte)object.asnType.tag;   // Tag
    return index;
  }

  /**
   * Return the OID subid as a numerical value
   */
  private long oidSubId(String[] subid, int index) throws CodingException {
    // TBD -- interface to MIB
    if (index >= subid.length) return 0;
    return Long.parseLong(subid[index]);
  }

  /**
   * Encode a sequence
   */
  int encodeSequence(SEQUENCE sequence, byte[] coding, int index)
    throws CodingException {
    int end = index;

    for (int i = sequence.contentCount - 1; i >= 0; i--)
      index = encode(sequence.contentData[i], coding, index);
    index = encodeLength(end - index, coding, index);
    coding[--index] = (byte)sequence.asnType.tag; // Tag
    return index;
  }

  /**WP
   * Encode a Set
   */
  int encodeSet(SEQUENCE sequence, byte[] coding, int index)
    throws CodingException {
    int end = index;

    for (int i = sequence.contentCount - 1; i >= 0; i--)
      index = encode(sequence.contentData[i], coding, index);
    index = encodeLength(end - index, coding, index);
    coding[--index] = (byte)sequence.asnType.tag; // Tag
    return index;
  }

  /**WP
   * Encode a context specific object
   */
  int encodeContextSpecific(CONTEXT context, byte[] coding, int index)
    throws CodingException {

    int tag = 0;
    int end = index;

    if (context.primitive) {                           // primitive
        tag = 0x80;

        byte[] data = (byte[])context.value.getValue(); // the primitive value
        System.arraycopy(data, 0, coding, index, data.length);
        index -= data.length;
    }
    else {                                              // constructed
        index = encode(context.value, coding, index);
        tag = 0xa0;
    }

    index = encodeLength(end - index, coding, index);
    coding[--index] = (byte)(tag | context.tag);       // Tag

    return index;
  }


  /****************************************************************
   * Decoding interface
   ****************************************************************/

  /**
   * Creates an ASN.1 object from the BER encoded version
   */
  public ASN1Object decode(byte[] coding) throws CodingException {
    int[] index = new int[1];
    index[0] = 0;

    return decode(coding, index);
  }

  /**
   * Creates an ASN.1 object from the BER encoded version
   * starting at given location.
   * @param coding A byte array containing the coded version
   * @param index A integer array of length one, initially containing the
   *              index of the first by of encoded version,
   *              and updated to point just one byte past
   *              the decoded data
   * @return a new OBJECT, created by reading the encoded version
   */
  ASN1Object decode(byte[] coding, int[] indexp) throws CodingException {
    int  index;			// Current reading index
    int  tag;			// Tag of object under decoding
    int  specificTag=0; // Tag of object when CONTEXT_SPECIFIC
    int  length;		// Length of data
    ASN1Object o;			// Create object

    index = indexp[0];
    tag = coding[index++] & 0xff;
    length = decodeLength(coding, index, indexp);
    index = indexp[0];

    try {
      if ((tag & 0xc0) == 0x80) {           // context specific
        specificTag = tag;
        tag = 0x80;
      }

      o = ASN.create(new ASN(tag, "Decoded value"));
      switch (tag) {
      // Generic ASN.1
      case 1:    decodeBoolean (o, coding, index, length); break;
      case 2:    decodeBigInt  (o, coding, index, length); break;
      case 5:    decodeNull    (o, coding, index, length); break;
      case 6:    decodeOID     (o, coding, index, length); break;
      case 0x30: decodeSequence(o, coding, index, length); break;
      // by Wolfgang Platzer
      case 3:    decodeBitString        (o, coding, index, length); break;//WP
      case 4:    decodeOctetString      (o, coding, index, length); break;//WP
      case 19:   decodePrintableString  (o, coding, index, length); break;//WP
      case 20:   decodeT61String        (o, coding, index, length); break;//WP
      case 22:   decodeIA5String        (o, coding, index, length); break;//WP
      case 23:   decodeUTCTime          (o, coding, index, length); break;//WP
      case 0x31: decodeSet              (o, coding, index, length); break;//WP
      case 0x80: decodeContextSpecific  (o, coding, index, length, specificTag); break;//WP

      case 7:    throw new InternalError("ObjectDesc coding not implemented");
      case 8:    throw new InternalError("External ASN.1 data not supported");
      case 9:    throw new InternalError("Real coding not implemented");
      case 10:   throw new InternalError("Enumerated coding not implemented");


      // Unrecognized
      default:
	throw new CodingException("Unrecognized ASN.1 TAG " + tag);
      }
    } catch (InstantiationException e) {
      throw new CodingException("ASN.1 creation error:" + e.getMessage());
    }
    indexp[0] += length;
    return o;
  }

  /**
   * Decode element length
   */
  int decodeLength(byte[] coding, int index, int[] end) throws CodingException {
      return decodeASN1Length(coding, index, end);
  }

  public static int decodeASN1Length(byte[] coding, int index, int[] end) throws CodingException {
    int firstbyte;
    int length;

    firstbyte = ((int)coding[index++]) & 0xff;

    if (firstbyte <= 127) {
      end[0] = index;
      return firstbyte;
    }

    if (firstbyte == 128)
      throw new CodingException("Indefinite format not supported.");

    firstbyte &= 0x7f;
    if (firstbyte > 4)
      throw new CodingException("Too large ASN.1 object");

    length = 0;
    while (--firstbyte >= 0)
      length = (length << 8) | (coding[index++] & 0xff);

    end[0] = index;
    return length;
  }

  /**
   * Decode a Boolean
   */
  void decodeBoolean(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    if (length != 1)
      throw new CodingException("Boolean length not one, length=" + length);

    o.setValue((coding[index++] == 0)? Boolean.FALSE: Boolean.TRUE);
  }

  /**WP
   * Decode an integer
   */
  void decodeBigInt(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    byte[] data = new byte[length];
    System.arraycopy(coding, index, data, 0, length);

    o.setValue(new BigInt(data));
  }

  /**WP @check
   * Decode a BitString
   */
  void decodeBitString(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    if (coding[index] != 0)
//      throw new CodingException("First Byte Of BitString not 0");
        System.out.println(coding[index]+" Bits are not valid in BitString!");

    byte[] data = new byte[length-1];

    System.arraycopy(coding, index+1, data, 0, length-1);

    o.setValue(data);
  }
  /**WP
   * Decode a OctetString
   */
  void decodeOctetString(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    byte[] data = new byte[length];

    System.arraycopy(coding, index, data, 0, length);

    o.setValue(data);
  }

  /**
   * Decode an Unsigned
   */
  void decodeUnsigned(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {
    int v;

    if (length > 5)
      throw new CodingException("Too long integer, length=" +length);
    if (coding[index] < 0)
      throw new CodingException("Signed integer when expecting unsigned");

    v = 0;
    while (--length >= 0)
      v = (v << 8) | (coding[index++] & 0xff);
    o.setValue(new Integer(v));
  }

  /**
   * Decode a string
   */
  void decodeString(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    o.setValue(new String(coding, 0, index, length));
  }

  /**WP
   * Decode a PrintableString
   */
  void decodePrintableString(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    o.setValue(new String(coding, 0, index, length));
  }

  /**WP
   * Decode a T61String
   */
  void decodeT61String(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    o.setValue(new String(coding, 0, index, length));
  }

  /**WP
   * Decode a IA5String
   */
  void decodeIA5String(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    o.setValue(new String(coding, 0, index, length));
  }

  /**WP
   * Decode a UTCTime
   */
  void decodeUTCTime(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    o.setValue(new String(coding, 0, index, length));
  }

  /**
   * Decode a null
   */
  void decodeNull(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {

    o.setValue(ASN.Null);
  }

  /** changed WP
   * Decode an OID.  OIDs are internally represented as Strings.
   */
  void decodeOID(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {
    StringBuffer buffer;	// Buffer to decode value into

    buffer = new StringBuffer();
    // First byte is special
    if (--length >= 0) {
      int byt    = coding[index++] & 0xff;
      int first  = (byt < 40? 0: (byt < 80? 1: 2));
      int second = (byt - first*40);
      buffer.append(first);
      buffer.append(".");
      buffer.append(second);
    }

    // Handle the rest
    while (length > 0) {
      buffer.append(".");

      int sid = 0;			// Subid
      do {
    	sid = (sid << 7) | (coding[index] & 0x7f);                  //WP bug
      } while (--length > 0 && (coding[index++] & 0x80) == 0x80);

      buffer.append(sid);
    }

    o.setValue(new String(buffer));
  }

  /**
   * Decode a sequence
   */
  void decodeSequence(ASN1Object o, byte[] coding, int index, int length)
    throws CodingException {
    int[] indexp = new int[1];

    indexp[0] = index;
    while (indexp[0] < index + length)
      o.addComponent(decode(coding, indexp));
  }

  /**WP
   * Decode a set as sequence
   */
  void decodeSet(ASN1Object o, byte[] coding, int index, int length)
  throws CodingException {
    int[] indexp = new int[1];

    indexp[0] = index;
    while (indexp[0] < index + length)
      o.addComponent(decode(coding, indexp));
  }

    /**WP
     * Decode a context specific object
     */
    void decodeContextSpecific(ASN1Object o, byte[] coding, int index, int length, int tag)
    throws CodingException {

        int[] indexp = new int[1];

        indexp[0] = index;

        byte[] data = new byte[length];
        System.arraycopy(coding, index, data, 0, length);

        CONTEXT con = null;
        if ((tag & 0x20) > 0)                           // constructed
            con = new CONTEXT(tag, decode(coding, indexp));
        else {                                          // primitive, put the value into a OCTET_STRING
            try {
                con = new CONTEXT(tag, ASN.create(ASN.OCTET_STRING, data));
            } catch (InstantiationException ex) {
                        // This line should never be displayed
                System.out.println("Error in decodeContextSpecific!");
            }
        }

        o.setValue(con);
    }

}
