// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.asn1;

import cryptiX.utils.ArrayEnumeration;
import cryptiX.utils.Composite;

import cryptiX.asn1.ASN;
import cryptiX.asn1.ASN1Object;

/**
 * This class represents ASN.1 context specific objects.
 * Note that sequences are NOT public.  They shall be
 * created with ASN.create(ASN.CONTEXT_SPECIFIC),
 * and accessed through the public OBJECT interface.
 */

final public class CONTEXT extends ASN1Object {
    /*
     * Initialize the object factory method.
     */
    static {
        Class _CONTEXT = new CONTEXT().getClass();
        ASN.register(ASN.CONTEXT_SPECIFIC, _CONTEXT);
    }

    /**
     * The context specific tag number of the object.
     */
    protected int tag;

    /**
     * shows if the encoding is primitive or constructed
     */
    protected boolean primitive;
    /**
     * The value of the context specific object
     */
    protected ASN1Object value;

    /**
     * Create a new CONTEXT.  Called by the OBJECTFactory.
     */
    protected CONTEXT() {
        tag = 0;
        primitive = false;
        value = null;
    }

    /**
     * Create a new CONTEXT.  Called by the OBJECTFactory.
     *
     * @param cst the context specific tag
     * @param the object defined by this tag
     */
    public CONTEXT(int cst, ASN1Object val) {
        tag = cst & 0x1f;
        if ((cst & 0x20) > 0)
            primitive = false;
        else primitive = true;
        value = val;
    }

    /**
     * Perform CONTEXT specific initialization
     */
    protected void init(ASN type) {
        this.asnType = type;
    }

    /**
     * This method is used to set the value of the context.
     *
     * @param cst the context specific tag
     * @param the object defined by this tag
     */
    public void setValue(int cst, ASN1Object val) {
        tag = cst & 0x1f;
        if ((cst & 0x20) > 0)
            primitive = false;
        else primitive = true;
        value = val;
    }

  /**
   * Sets the object value.
   */
    public synchronized void setValue(Object obj) {
        CONTEXT con = (CONTEXT)obj;
        tag = con.tag;
        primitive = con.primitive;
        value = con.value;
    }

    /**
     * Returns the tag of this context
     */
    public int tag() {
        return tag;
    }

    /**
     * Returns the value of this context
     */
    public Object getValue() {
        return value;
    }

    /**
     * Returns the value of this context
     */
    public boolean isPrimitive() {
        return primitive;
    }

    /**
     * Returns a String that represents the value of this Object.
     */
    public String toString() {

//        String str = value.getClass().getName();

        return asnType.getName()+" = "+"["+getClass().getName()+"] = tag: "+tag+" ["+"]";
    }
}
