// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.rsa;

import java.io.*;
import java.util.*;

import cryptiX.asn1.CodingException;

/**
 * Class for handling private keys
 *
 */
public class PrivateKey extends PublicKey implements Observer {

    BigInt privateExponent = null;         // private exponent
    BigInt prime1 = null;                  // one factor of n
    BigInt prime2 = null;                  // second factor of n
    BigInt exponent1 = null;               // d mod (p1-1)
    BigInt exponent2 = null;               // d mod (p2-1)
    BigInt coefficient = null;             // multiplic inverse of p mod q

    /**
     * The default constructor.
     */
    public PrivateKey() {
    }

    /**
     * Encrypt or decrypt a BigInt using the private key.
     * Simply performs a modula exponentiation.
     *
     * @param message the BigInt to de/encrypt
     */
    public BigInt cryptPrivate(BigInt message) {

        return RSAFunctions.mExp(message, privateExponent, modulus);
    }

    public void update(Observable o, Object arg) {
        setChanged();
        notifyObservers(arg);
    }

    /**
     * generates the key pair and uses the Java default random number generator
     *
     * @param keylen number of bits for modulus
     */
    public void generateKeys(int keylen) {
        generateKeys(keylen, null);
    }

    /**
     * generates the key pair and uses a give random number generator
     *
     * @param keylen number of bits for modulus
     * @param rg the random number generator to use
     */

    public void generateKeys(int keylen, java.util.Random rg) {

        int p1bits = keylen/2;
        int p2bits;

        RandomPrimeGenerator rpg = new RandomPrimeGenerator(rg);

        rpg.addObserver(this);

        setChanged();
        notifyObservers("Generating RSA Key Pair...");

        setChanged();
        notifyObservers("Generating first prime...");
        prime1 = rpg.randomPrime(p1bits);

        p2bits = keylen - prime1.bitLength();
        setChanged();
        notifyObservers("Generating second prime...");
        prime2 = rpg.randomPrime(p2bits);

   		if (prime2.lessThan(prime1)) {            // ensure that p1<p2 for computing u
			coefficient = prime1;
			prime1 = prime2;
			prime2 = coefficient;
		}

        BigInt p1_1 = prime1.subtract(BigInt.ONE);
        BigInt p2_1 = prime2.subtract(BigInt.ONE);

        BigInt phi = BigInt.multiply(p1_1, p2_1);

        publicExponent = BigInt.ONE;    // searching encryption key
                                        // start with 3, till gcd(e, phi) = 1
        do {
            publicExponent = publicExponent.add(BigInt.TWO);
        } while (RSAFunctions.gcd(publicExponent, phi).compare(BigInt.ONE) != 0);


        modulus = prime1.multiply(prime2);

        privateExponent = RSAFunctions.mulInv(publicExponent, phi);

        exponent1 = privateExponent.mod(p1_1);
        exponent2 = privateExponent.mod(p2_1);

        coefficient = RSAFunctions.mulInv(prime1, prime2);

        setChanged();
        notifyObservers("Key Generation OK!");
    }

    /**
     * returns the public key part of this private key
     */
    public PublicKey publicKey() {

        return new PublicKey(ID, publicExponent, modulus);
    }

    /**
     * creates a signature according to PKCS #1
     */
    public byte[] createSignature(byte[] message, String signatureAlgorithm)
        throws CodingException {

        if (signatureAlgorithm.equals(RSAFunctions.md2WithRSAEncryption))
            return RSAFunctions.createSignature(message, this, RSAFunctions.md2);
        else if (signatureAlgorithm.equals(RSAFunctions.md4WithRSAEncryption))
            return RSAFunctions.createSignature(message, this, RSAFunctions.md4);
        else if (signatureAlgorithm.equals(RSAFunctions.md5WithRSAEncryption))
            return RSAFunctions.createSignature(message, this, RSAFunctions.md5);
        else if (signatureAlgorithm.equals(RSAFunctions.sha1WithRSAEncryption))
            return RSAFunctions.createSignature(message, this, RSAFunctions.sha1);
        else                                            // only RSA at this time
            new CodingException("Signature Algorithm "+signatureAlgorithm+" not implemented!");

        return null;
    }

    /**
     * returns the private exponent
     */
    public BigInt privateExponent() {
        return new BigInt(privateExponent);
    }

    /**
     * returns the first prime
     */
    public BigInt prime1() {
        return new BigInt(prime1);
    }

    /**
     * returns the second prime
     */
    public BigInt prime2() {
        return new BigInt(prime2);
    }

    /**
     * returns the first exponent
     */
    public BigInt exponent1() {
        return new BigInt(exponent1);
    }

    /**
     * returns the second exponent
     */
    public BigInt exponent2() {
        return new BigInt(exponent2);
    }

    /**
     * returns the coefficient
     */
    public BigInt coefficient() {
        return new BigInt(coefficient);
    }

    public String toString() {
        String str = super.toString();

        str = str.concat("private exponent: "+privateExponent+"\n");
        str = str.concat("prime 1:          "+prime1+"\n");
        str = str.concat("prime 2:          "+prime2+"\n");
        str = str.concat("exponent 1:       "+exponent1+"\n");
        str = str.concat("exponent 2:       "+exponent2+"\n");
        str = str.concat("coefficient:      "+coefficient+"\n");

        return str;
    }
}