// Copyright (C) 1996 by Wolfgang Platzer
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//

package cryptiX.user;

import java.io.*;
import java.util.Date;
import cryptiX.asn1.*;
import cryptiX.rsa.*;
import cryptiX.crypto.Md5Hash;
import cryptiX.utils.*;

/**
 * This class represents a X509 certificate request.
 */

public class X509CertRequest {

    private ASN1 x509CertRequest = null;        // certificate according to X.509
    private ASN1 certRequestInfo = null;        // certificate info according to X.509

    private int saveFormat = ASN1.DER;          // default save format is DER

    private int         version = 0;                    // version number
    private Name        subject = null;                 // name of the subject
    private PublicKey   publicKey = null;               // the public inside this certificate
    private String      signatureAlgorithmID = null;    // signature algorithm id

    /**
     * create a x509CertRequest form a PEM or DER file
     */
    public X509CertRequest(String fileName) throws IOException, CodingException {
        x509CertRequest = new ASN1(fileName);
        saveFormat = x509CertRequest.saveFormat();

        if (x509CertRequest != null)
            init();
    }

    /**
     * create a x509CertRequest form a PEM or DER array
     */
    public X509CertRequest(byte[] arr) throws CodingException {
        x509CertRequest = new ASN1(arr);

        if (x509CertRequest != null)
            init();
    }

    /**
     * default constructor
     */
    public X509CertRequest() {
        x509CertRequest = new ASN1();
    }

    /**
     * inits the data from the x509 certificate request
     */
    private void init() throws CodingException {
        certRequestInfo = new ASN1(x509CertRequest.getComponentAt(0));
        version = ((BigInt)certRequestInfo.getComponentAt(0).getValue()).intValue();
        subject = new Name(certRequestInfo.getComponentAt(1));

        ASN1Object spki = certRequestInfo.getComponentAt(2);
        publicKey = PublicKey.valueOf(spki);

        if (certRequestInfo.components() > 3) {     // Extensions

            ASN1Object obj = certRequestInfo.getComponentAt(3);
            ASN1Object obj1 = (ASN1Object)obj.getValue();

//            System.out.println(obj);
//            System.out.println(new ASN1(obj1));
        }
    }

    /**
     * creates the ASN1 object from the variables
     */
    private void createCertRequestInfo() {

        ASN1Object cri = null;           // certificate info

        try {
            cri = ASN.create(ASN.SEQUENCE);

            cri.addComponent(ASN.create(ASN.INTEGER, new BigInt((long)version)));
            cri.addComponent(subject.toASN1Object());
            cri.addComponent(publicKey.toSubjectPublicKeyInfo());

            // @todo context specific
        } catch (InstantiationException ex) {
        }
        certRequestInfo = new ASN1(cri);
    }

    /**
     * signs the certificate with the private key of the issuer
     *
     * @param issuerSK the private key of the issuer
     */
    public void signCertRequest(PrivateKey issuerSK) throws CodingException {

        createCertRequestInfo();

        byte[] signature = issuerSK.createSignature(certRequestInfo.toByteArrayDER(), signatureAlgorithmID);

        ASN1Object c = null;

        try {
            c = ASN.create(ASN.SEQUENCE);
            c.addComponent(certRequestInfo.toASN1Object());
            c.addComponent(Util.AlgID(signatureAlgorithmID));
            c.addComponent(ASN.create(ASN.BIT_STRING, signature));
        } catch (InstantiationException ex) {
        }

        x509CertRequest = new ASN1(c);
    }

    /**
     * verifies the self signed certificate request
     *
     * @param signerPK the public key from the signer
     * @return true if signature OK
     */
    public boolean verify() throws CodingException {

                        // get the signatureAlgorithm
        String signatureAlgorithm = Util.AlgIDAlgorithm(x509CertRequest.getComponentAt(1));

                        // get the signature from the certificate
        ASN1Object sigObject = x509CertRequest.getComponentAt(2);
        byte[] signature = (byte[])(sigObject).getValue();


        byte[] signedData = x509CertRequest.getFirstObject();

        return publicKey.checkSignature(signedData, signature, signatureAlgorithm);

                        // check the signature with appropriate algorithm
//        return signerPK.checkSignature(certificateInfo.toByteArray(), signature, signatureAlgorithm);

    }

    /**
     * returns the PEM format of the certificate request
     */
    public byte[] toPEM() {
        return x509CertRequest.toByteArrayPEM();
    }

    /**
     * returns the DER format of the certificate request
     */
    public byte[] toDER() {
        return x509CertRequest.toByteArrayDER();
    }

    /**
     * saves the certificate request to a file
     *
     * @param fileName name of the file
     * @param pemFormat true, if file format should be PEM
     *                  false, if file format should be DER
     */
    public void saveToFile(String fileName) throws IOException {

        if (saveFormat == ASN1.PEM)
            x509CertRequest.saveToFile(fileName, "-----BEGIN CERTIFICATE-----",
                "-----END CERTIFICATE-----");
        else
            x509CertRequest.saveToFile(fileName);
    }

    /**
     * converts the certificate request to a text output
     */
    public String toString() {
        return x509CertRequest.toString();
    }

    /**
     * sets the file format (ASN1.DER, ASN1.PEM) of the cert request
     */
    public void setSaveFormat(int format) {
        saveFormat = format;
    }

    /**
     * sets the version of this certificate request
     */
    public void setVersion(int v) {
        version = v;
    }

    /**
     * sets the signature algorithm of this certificate request
     */
    public void setSignatureAlgorithmID(String said) {
        signatureAlgorithmID = said;
    }

    /**
     * sets the subject of this certificate request
     */
    public void setSubject(Name su) {
        subject = su;
    }

    /**
     * sets the public key of this certificate request
     */
    public void setPublicKey(PublicKey pk) {
        publicKey = pk;
    }

    /**
     * returns the file format (ASN1.DER, ASN1.PEM) of the cert request
     */
    public int saveFormat() {
        return saveFormat;
    }

    /**
     * returns the version of this certificate request
     */
    public int version() {
        return version;
    }

    /**
     * returns the signature algorithm of this certificate request
     */
    public String signatureAlgorithmID() {
        return signatureAlgorithmID;
    }

    /**
     * returns the subject of this certificate request
     */
    public Name subject() {
        return subject;
    }

    /**
     * returns the public key of this certificate request
     */
    public PublicKey publicKey() {
        return publicKey;
    }
}
